import numba
import numpy as np
from immutabledict import immutabledict

import strax
import straxen
export, __all__ = strax.exporter()


MV_PREAMBLE = 'Muno-Veto Plugin: Same as the corresponding nVETO-PLugin.\n'

@export
@strax.takes_config(
    strax.Option(
        'save_outside_hits_nv',
        default=(3, 15), track=True,
        help='Save (left, right) samples besides hits; cut the rest'),
    strax.Option(
        'hit_min_amplitude_nv',
        default=20, track=True,
        help='Minimum hit amplitude in ADC counts above baseline. '
             'Specify as a tuple of length n_nveto_pmts, or a number.'),
    strax.Option(
        'min_split_nv',
        default=100, track=True,
        help='Minimum height difference pe/sample between local minimum and maximum, '
             'that a pulse get split.'),
    strax.Option(
        'min_split_ratio_nv',
        default=0, track=True,
        help='Min ratio between local maximum and minimum to split pulse (zero to switch this off).'),
    strax.Option(
        'entropy_template_nv',
        default='flat', track=True,
        help='Template data is compared with in conditional entropy. Can be either "flat" or an template array.'),
    strax.Option(
        'entropy_square_data_nv',
        default=False, track=True,
        help='Parameter which decides if data is first squared before normalized and compared to the template.'),
    strax.Option('channel_map', track=False, type=immutabledict,
                 help="immutabledict mapping subdetector to (min, max) "
                      "channel number."),
    strax.Option('n_nveto_pmts', type=int, track=False,
                 help='Number of muVETO PMTs'),
    strax.Option('gain_model_nv',
             help='PMT gain model. Specify as (model_type, model_config)'),
)
class nVETOHitlets(strax.Plugin):
    """
    Plugin which computes the nveto hitlets and their parameters.
    
    Hitlets are an extension of regular hits. They include the left
    and right extension. The plugin does the following:
        1. Generate hitlets which includes these sub-steps:
            * Apply left and right hit extension and concatenate
            overlapping hits.
            * Generate temp. hitelts and look for their waveforms in
            their corresponding records.
            * Split hitlets if they satisfy the set criteria.
        2. Compute the properties of the hitlets.

    Note:
        Hitlets are getting chopped if extended in not recorded regions.
    """
    __version__ = '0.0.6'

    parallel = 'process'
    rechunk_on_save = True
    compressor = 'zstd'

    depends_on = 'records_nv'

    provides = 'hitlets_nv'
    data_kind = 'hitlets_nv'
    ends_with = '_nv'

    dtype = strax.hitlet_dtype()

    def setup(self):
        to_pe = straxen.get_to_pe(self.run_id,
                                  self.config['gain_model_nv'],
                                  self.config['n_nveto_pmts'])
        self.channel_range = self.config['channel_map']['nveto']
        
        # Create to_pe array of size max channel:
        self.to_pe = np.zeros(self.channel_range[1] + 1, dtype=np.float32)
        self.to_pe[self.channel_range[0]:] = to_pe[:]

    def compute(self, records_nv, start, end):
        # Search again for hits in records:
        hits = strax.find_hits(records_nv, min_amplitude=self.config['hit_min_amplitude_nv'])
        # Merge concatenate overlapping  within a channel. This is important
        # in case hits were split by record boundaries. In case we
        # accidentally concatenate two PMT signals we split them later again.
        hits = strax.concat_overlapping_hits(hits,
                                             self.config['save_outside_hits_nv'],
                                             self.channel_range,
                                             start,
                                             end)
        hits = strax.sort_by_time(hits)

        # Now convert hits into temp_hitlets including the data field:
        if len(hits):
            nsamples = hits['length'].max()
        else:
            nsamples = 0
        temp_hitlets = np.zeros(len(hits), strax.hitlet_with_data_dtype(n_samples=nsamples))
    
        # Generating hitlets and copying relevant information from hits to hitlets.
        # These hitlets are not stored in the end since this array also contains a data
        # field which we will drop later.
        strax.refresh_hit_to_hitlets(hits, temp_hitlets)
        del hits
        
        # Get hitlet data and split hitlets:
        strax.get_hitlets_data(temp_hitlets, records_nv, to_pe=self.to_pe)

        temp_hitlets = strax.split_peaks(temp_hitlets,
                                         records_nv,
                                         self.to_pe,
                                         data_type='hitlets',
                                         algorithm='local_minimum',
                                         min_height=self.config['min_split_nv'],
                                         min_ratio=self.config['min_split_ratio_nv']
                                         )

        # Compute other hitlet properties:
        # We have to loop here 3 times over all hitlets...
        strax.hitlet_properties(temp_hitlets)
        entropy = strax.conditional_entropy(temp_hitlets, template='flat', square_data=False)
        temp_hitlets['entropy'][:] = entropy

        # Remove data field:
        hitlets = np.zeros(len(temp_hitlets), dtype=strax.hitlet_dtype())
        strax.copy_to_buffer(temp_hitlets, hitlets, '_copy_hitlets')
        return hitlets


@export
@strax.takes_config(
    strax.Option(
        'save_outside_hits_mv',
        default=(2, 5), track=True,
        child_option=True, parent_option_name='save_outside_hits_nv',
        help='Save (left, right) samples besides hits; cut the rest'),
    strax.Option(
        'hit_min_amplitude_mv',
        default=20, track=True,
        child_option=True, parent_option_name='hit_min_amplitude_nv',
        help='Minimum hit amplitude in ADC counts above baseline. '
             'Specify as a tuple of length n_nveto_pmts, or a number.'),
    strax.Option(
        'min_split_mv', 
        default=100, track=True,
        child_option=True, parent_option_name='min_split_nv',
        help='Minimum height difference pe/sample between local minimum and maximum, '
             'that a pulse get split.'),
    strax.Option(
        'min_split_ratio_mv',
        default=0, track=True,
        child_option=True, parent_option_name='min_split_ratio_nv',
        help='Min ratio between local maximum and minimum to split pulse (zero to switch this off).'),
    strax.Option(
        'entropy_template_mv',
        default='flat', track=True,
        child_option=True, parent_option_name='entropy_template_nv',
        help='Template data is compared with in conditional entropy. Can be either "flat" or an template array.'),
    strax.Option(
        'entropy_square_data_mv',
        default=False, track=True,
        child_option=True, parent_option_name='entropy_square_data_nv',
        help='Parameter which decides if data is first squared before normalized and compared to the template.'),
    strax.Option('gain_model_mv',
                 child_option=True, parent_option_name='gain_model_nv',
             help='PMT gain model. Specify as (model_type, model_config)'),
    strax.Option('n_mveto_pmts', type=int, track=False,
                 help='Number of muVETO PMTs'),
)
class muVETOHitlets(nVETOHitlets):
    __doc__ = MV_PREAMBLE + nVETOHitlets.__doc__
    __version__ = '0.0.2'
    depends_on = 'records_mv'

    provides = 'hitlets_mv'
    data_kind = 'hitlets_mv'
    child_plugin = True

    dtype = strax.hitlet_dtype()

    def setup(self):
        to_pe = straxen.get_to_pe(self.run_id,
                                  self.config['gain_model_mv'],
                                  straxen.n_mveto_pmts)
        self.channel_range = self.config['channel_map']['mv']
        
        # Create to_pe array of size max channel:
        self.to_pe = np.zeros(self.channel_range[1] + 1, dtype=np.float32)
        self.to_pe[self.channel_range[0]:] = to_pe[:]

    def compute(self, records_mv, start, end):
        return super().compute(records_mv, start, end)



import numpy as np
import strax
import straxen
from warnings import warn

export, __all__ = strax.exporter()
__all__ += ['FIXED_TO_PE']


@export
def get_to_pe(run_id, gain_model, n_pmts):
    if not isinstance(gain_model, tuple):
        raise ValueError(f"gain_model must be a tuple")
    if not len(gain_model) == 2:
        raise ValueError(f"gain_model must have two elements: "
                         f"the model type and its specific configuration")
    model_type, model_conf = gain_model

    if model_type == 'disabled':
        # Somebody messed up
        raise RuntimeError("Attempt to use a disabled gain model")
    if model_type == 'CMT_model':
        if not isinstance(model_conf, tuple) or len(model_conf) != 2:
            # Raise a value error if the condition is not met. We should have:
            # ("CMT_model", -> To specify that we want to use the online
            #                  corrections management tool
            #   (
            #   "to_pe_model", -> This is to specify that we want  the gains
            #   "v1", -> The version of the correction 'v1' is the online version
            #   )
            raise ValueError('CMT gain model should be similar to:'
                             '("CMT_model", ("to_pe_model", "v1"). Instead got:'
                             f'{model_conf}')
        # is this the best way to do this?
        is_nt = n_pmts == straxen.n_tpc_pmts or n_pmts == straxen.n_nveto_pmts or n_pmts == straxen.n_mveto_pmts

        corrections = straxen.CorrectionsManagementServices(is_nt=is_nt)
        to_pe = corrections.get_corrections_config(run_id, model_conf)

        return to_pe

    elif model_type == 'to_pe_per_run':
        warn("to_pe_per_run will be replaced by CorrectionsManagementSevices",
             DeprecationWarning, 2)
        # Load a npy file specifying a run_id -> to_pe array
        to_pe_file = model_conf
        x = straxen.get_resource(to_pe_file, fmt='npy')
        run_index = np.where(x['run_id'] == int(run_id))[0]
        if not len(run_index):
            # Gains not known: using placeholders
            run_index = [-1]
        to_pe = x[run_index[0]]['to_pe']

    elif model_type == 'to_pe_constant':
        if model_conf in FIXED_TO_PE:
            return FIXED_TO_PE[model_conf]

        try:
            # Uniform gain, specified as a to_pe factor
            to_pe = np.ones(n_pmts, dtype=np.float32) * model_conf
        except np.core._exceptions.UFuncTypeError as e:
            raise (str(e) +
                   f'\nTried multiplying by {model_conf}. Insert a number instead.')
    else:
        raise NotImplementedError(f"Gain model type {model_type} not implemented")

    if len(to_pe) != n_pmts:
        raise ValueError(
            f"Gain model {gain_model} resulted in a to_pe "
            f"of length {len(to_pe)}, but n_pmts is {n_pmts}!")
    return to_pe


FIXED_TO_PE = {
    # just some dummy placeholder for nT gains
    'gain_placeholder': np.repeat(0.0085, straxen.n_tpc_pmts),
    # Gains which will preserve all areas in adc counts.
    # Useful for debugging and tests.
    'adc_tpc': np.ones(straxen.n_tpc_pmts),
    'adc_mv': np.ones(straxen.n_mveto_pmts),
    'adc_nv': np.ones(straxen.n_nveto_pmts)
}


@export
def get_elife(run_id, elife_conf):
    if isinstance(elife_conf, tuple) and len(elife_conf) == 3:
        # We want to use corrections management
        is_nt = elife_conf[-1]
        cmt = straxen.CorrectionsManagementServices(is_nt=is_nt)

        e = cmt.get_corrections_config(run_id, elife_conf[:2])

    elif isinstance(elife_conf, str):
        warn("get_elife will be replaced by CorrectionsManagementSevices",
             DeprecationWarning, 2)
        # Let's remove these functions and only rely on the CMT in the future
        x = straxen.get_resource(elife_conf, fmt='npy')
        run_index = np.where(x['run_id'] == int(run_id))[0]
        if not len(run_index):
            # Electron lifetime not known: using placeholders
            e = 623e3
        else:
            e = x[run_index[0]]['e_life']
    else:
        raise ValueError(
            'Wrong elife model. Either specify a string (url) or the '
            'Corrections Management Tools format: '
            '(model_type->str, model_config->str, is_nT->bool)'
            '')
    return float(e)


@export
def get_config_from_cmt(run_id, conf):
    if isinstance(conf, str) and conf.startswith('https://raw'):
        # Legacy support for pax files
        return conf
    if not isinstance(conf, tuple):
        raise ValueError(f'conf must be a tuple')
    if not len(conf) == 3:
        raise ValueError(f'conf must have three elements: '
                         f'the model type, its specific configuration '
                         f'and detector (True = nT)')
    model_type, model_conf, is_nt = conf
    if model_type == 'CMT_model':
        cmt = straxen.CorrectionsManagementServices(is_nt=is_nt)
        this_file = cmt.get_corrections_config(run_id, model_conf)
        this_file = ' '.join(map(str, this_file))

    else:
        raise ValueError(f'Wrong NN configuration, please look at this {nn_config} '
                         'and modify it accordingly')

    return this_file



__version__ = '0.15.6'

from utilix import uconfig
from .common import *
# contexts.py below
from .corrections_services import *
from .get_corrections import *
from .hitfinder_thresholds import *
from .itp_map import *
from .matplotlib_utils import *
from .mini_analysis import *
from .misc import *
from .mongo_storage import *
from .online_monitor import *
from .rundb import *
from .scada import *
from .bokeh_utils import *


from . import plugins
from .plugins import *

from . import analyses

# Do not make all contexts directly available under straxen.
# Otherwise we have straxen.demo() etc.
from . import contexts



from immutabledict import immutabledict
import strax
import straxen


common_opts = dict(
    register_all=[
        straxen.event_processing,
        straxen.double_scatter],
    # Register all peak/pulse processing by hand as 1T does not need to have
    # the high-energy plugins.
    register=[
        straxen.PulseProcessing,
        straxen.Peaklets,
        straxen.PeakletClassification,
        straxen.MergedS2s,
        straxen.Peaks,
        straxen.PeakBasics,
        straxen.PeakProximity],
    check_available=('raw_records', 'peak_basics'),
    store_run_fields=(
        'name', 'number',
        'start', 'end', 'livetime', 'mode'))

xnt_common_config = dict(
    n_tpc_pmts=straxen.n_tpc_pmts,
    n_top_pmts=straxen.n_top_pmts,
    n_mveto_pmts=straxen.n_mveto_pmts,
    n_nveto_pmts=straxen.n_nveto_pmts,
    gain_model=("CMT_model", ("to_pe_model", "ONLINE")),
    gain_model_nv=("to_pe_constant", "adc_nv"),
    gain_model_mv=("to_pe_constant", "adc_mv"),
    channel_map=immutabledict(
        # (Minimum channel, maximum channel)
        # Channels must be listed in a ascending order!
        tpc=(0, 493),
        he=(500, 752),  # high energy
        aqmon=(790, 807),
        aqmon_nv=(808, 815),  # nveto acquisition monitor
        tpc_blank=(999, 999),
        mv=(1000, 1083),
        aux_mv=(1084, 1087),  # Aux mv channel 2 empty  1 pulser  and 1 GPS
        mv_blank=(1999, 1999),
        nveto=(2000, 2119),
        nveto_blank=(2999, 2999)),
    # Clustering/classification parameters
    s1_max_rise_time=100,
    s2_xy_correction_map=("CMT_model", ('s2_xy_map', "ONLINE"), True),
    elife_file=("elife_model", "ONLINE",True),
)

# Plugins in these files have nT plugins, E.g. in pulse&peak(let)
# processing there are plugins for High Energy plugins. Therefore do not
# st.register_all in 1T contexts.
xnt_common_opts = common_opts.copy()
xnt_common_opts['register'] = common_opts['register'] + [
    straxen.PeakPositionsCNN,
    straxen.PeakPositionsMLP,
    straxen.PeakPositionsGCN,
    straxen.PeakPositionsNT,
    straxen.PeakBasicsHighEnergy,
    straxen.PeaksHighEnergy,
    straxen.PeakletsHighEnergy,
    straxen.PeakletClassificationHighEnergy,
    straxen.MergedS2sHighEnergy,
]
xnt_common_opts['register_all'] = common_opts['register_all'] + [
    straxen.nveto_recorder,
    straxen.veto_pulse_processing,
    straxen.veto_hitlets,
    straxen.acqmon_processing,
    straxen.pulse_processing,
    straxen.peaklet_processing,
    straxen.online_monitor,
]

##
# XENONnT
##


def xenonnt_online(output_folder='./strax_data',
                   we_are_the_daq=False,
                   _minimum_run_number=7157,
                   _database_init=True,
                   **kwargs):
    """XENONnT online processing and analysis"""
    context_options = {
        **straxen.contexts.xnt_common_opts,
        **kwargs}

    st = strax.Context(
        config=straxen.contexts.xnt_common_config,
        **context_options)
    st.register([straxen.DAQReader, straxen.LEDCalibration])

    st.storage = [straxen.RunDB(
        readonly=not we_are_the_daq,
        minimum_run_number=_minimum_run_number,
        runid_field='number',
        new_data_path=output_folder,
        rucio_path='/dali/lgrandi/rucio/')
        ] if _database_init else []
    if not we_are_the_daq:
        st.storage += [
            strax.DataDirectory(
                '/dali/lgrandi/xenonnt/raw',
                readonly=True,
                take_only=straxen.DAQReader.provides),
            strax.DataDirectory(
                '/dali/lgrandi/xenonnt/processed',
                readonly=True)]
        if output_folder:
            st.storage.append(
                strax.DataDirectory(output_folder))

        st.context_config['forbid_creation_of'] = straxen.daqreader.DAQReader.provides + ('records',)
    # Only the online monitor backend for the DAQ
    elif _database_init:
        st.storage += [straxen.OnlineMonitor(
            readonly=not we_are_the_daq,
            take_only=('veto_intervals',
                       'online_peak_monitor',
                       'event_basics',))]

    # Remap the data if it is before channel swap (because of wrongly cabled
    # signal cable connectors) These are runs older than run 8797. Runs
    # newer than 8796 are not affected. See:
    # https://github.com/XENONnT/straxen/pull/166 and
    # https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:dsg:daq:sector_swap
    st.set_context_config({'apply_data_function': (straxen.common.remap_old,)})
    return st


def xenonnt_led(**kwargs):
    st = xenonnt_online(**kwargs)
    st.context_config['check_available'] = ('raw_records', 'led_calibration')
    # Return a new context with only raw_records and led_calibration registered
    st = st.new_context(
        replace=True,
        config=st.config,
        storage=st.storage,
        **st.context_config)
    st.register([straxen.DAQReader, straxen.LEDCalibration])
    return st


# This gain model is the average to_pe. For something more fancy use the CMT
def xenonnt_simulation(output_folder='./strax_data'):
    import wfsim
    xnt_common_config['gain_model'] = ('to_pe_constant', 0.01)
    st = strax.Context(
        storage=strax.DataDirectory(output_folder),
        config=dict(detector='XENONnT',
                    fax_config='fax_config_nt.json',
                    check_raw_record_overlaps=False,
                    **straxen.contexts.xnt_common_config,
                    ),
        **straxen.contexts.xnt_common_opts)
    st.register(wfsim.RawRecordsFromFaxNT)
    return st


def xenonnt_temporary_five_pmts(**kwargs):
    """Temporary context for selected PMTs"""
    # Start from the online context
    st_online = xenonnt_online(**kwargs)

    temporary_five_pmts_config = {
        'gain_model': ('CMT_model', ("to_pe_model", "xenonnt_temporary_five_pmts")),
        'peak_min_pmts': 2,
        'peaklet_gap_threshold': 300,
    }
    # If there are any config overwrites in the kwargs, us those,
    # otherwise use the config as in the dict above.

    for k in list(temporary_five_pmts_config.keys()):
        if k in kwargs:
            temporary_five_pmts_config[k] = kwargs[k]

    # Copy the online context and change the configuration here
    st = st_online.new_context()
    st.set_config(temporary_five_pmts_config)

    return st


def xenonnt_initial_commissioning(*args, **kwargs):
    raise ValueError(
        'Use xenonnt_online. See' 
        'https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:analysis:commissioning:straxen_contexts#update_09_nov_20')  # noqa

##
# XENON1T
##


x1t_context_config = {
    **common_opts,
    **dict(
        check_available=('raw_records', 'records', 'peaklets',
                         'events', 'event_info'),
        free_options=('channel_map',),
        store_run_fields=tuple(
            [x for x in common_opts['store_run_fields'] if x != 'mode']
            + ['trigger.events_built', 'reader.ini.name']))}
x1t_context_config.update(
    dict(register=common_opts['register'] +
                  [straxen.PeakPositions1T,
                   straxen.RecordsFromPax,
         ]))

x1t_common_config = dict(
    check_raw_record_overlaps=False,
    allow_sloppy_chunking=True,
    n_tpc_pmts=248,
    n_top_pmts=127,
    channel_map=immutabledict(
        # (Minimum channel, maximum channel)
        tpc=(0, 247),
        diagnostic=(248, 253),
        aqmon=(254, 999)),
    hev_gain_model=('to_pe_per_run',
                    'to_pe.npy'),
    gain_model=('to_pe_per_run',
                'to_pe.npy'),
    pmt_pulse_filter=(
        0.012, -0.119,
        2.435, -1.271, 0.357, -0.174, -0., -0.036,
        -0.028, -0.019, -0.025, -0.013, -0.03, -0.039,
        -0.005, -0.019, -0.012, -0.015, -0.029, 0.024,
        -0.007, 0.007, -0.001, 0.005, -0.002, 0.004, -0.002),
    tail_veto_threshold=int(1e5),
    # Smaller right extension since we applied the filter
    peak_right_extension=30,
    peak_min_pmts=2,
    save_outside_hits=(3, 3),
    hit_min_amplitude='XENON1T_SR1',
    peak_split_gof_threshold=(
        None,  # Reserved
        ((0.5, 1), (3.5, 0.25)),
        ((2, 1), (4.5, 0.4))),
    left_event_extension=int(1e6),
    right_event_extension=int(1e6),
)



def demo():
    """Return strax context used in the straxen demo notebook"""
    straxen.download_test_data()

    st = strax.Context(
        storage=[strax.DataDirectory('./strax_data'),
                 strax.DataDirectory('./strax_test_data',
                                     deep_scan=True,
                                     provide_run_metadata=True,
                                     readonly=True)],
        forbid_creation_of=straxen.daqreader.DAQReader.provides,
        config=dict(**x1t_common_config),
        **x1t_context_config)

    # Use configs that are always available
    st.set_config(dict(
        hev_gain_model=('to_pe_per_run', straxen.aux_repo +
                        '3548132b55f81a43654dba5141366041e1daaf01/strax_files/to_pe.npy'),
        gain_model=('to_pe_per_run', straxen.aux_repo +
                    '3548132b55f81a43654dba5141366041e1daaf01/strax_files/to_pe.npy'),
    ))
    return st


def fake_daq():
    """Context for processing fake DAQ data in the current directory"""
    st = strax.Context(
        storage=[strax.DataDirectory('./strax_data'),
                 # Fake DAQ puts run doc JSON in same folder:
                 strax.DataDirectory('./from_fake_daq',
                                     provide_run_metadata=True,
                                     readonly=True)],
        config=dict(daq_input_dir='./from_fake_daq',
                    daq_chunk_duration=int(2e9),
                    daq_compressor='lz4',
                    n_readout_threads=8,
                    daq_overlap_chunk_duration=int(2e8),
                    **x1t_common_config),
        **x1t_context_config)
    st.register(straxen.Fake1TDAQReader)
    return st


def xenon1t_dali(output_folder='./strax_data', build_lowlevel=False, **kwargs):
    context_options = {
        **x1t_context_config,
        **kwargs}

    st = strax.Context(
        storage=[
            strax.DataDirectory(
                '/dali/lgrandi/xenon1t/strax_converted/raw',
                take_only='raw_records',
                provide_run_metadata=True,
                readonly=True),
            strax.DataDirectory(
                '/dali/lgrandi/xenon1t/strax_converted/processed',
                readonly=True),
            strax.DataDirectory(output_folder)],
        config=dict(**x1t_common_config),
        # When asking for runs that don't exist, throw an error rather than
        # starting the pax converter
        forbid_creation_of=(
            straxen.daqreader.DAQReader.provides if build_lowlevel
            else straxen.daqreader.DAQReader.provides + ('records', 'peaklets')),
        **context_options)
    return st


def xenon1t_led(**kwargs):
    st = xenon1t_dali(**kwargs)
    st.context_config['check_available'] = ('raw_records', 'led_calibration')
    # Return a new context with only raw_records and led_calibration registered
    st = st.new_context(
        replace=True,
        config=st.config,
        storage=st.storage,
        **st.context_config)
    st.register([straxen.RecordsFromPax, straxen.LEDCalibration])
    return st


def xenon1t_simulation(output_folder='./strax_data'):
    import wfsim
    st = strax.Context(
        storage=strax.DataDirectory(output_folder),
        config=dict(
            fax_config='fax_config_1t.json',
            detector='XENON1T',
            check_raw_record_overlaps=False,
            **straxen.contexts.x1t_common_config),
        **straxen.contexts.common_opts)
    st.register(wfsim.RawRecordsFromFax1T)
    return st



import strax
import numpy as np
import numexpr

export, __all__ = strax.exporter()


@export
@strax.takes_config(
    strax.Option(
        'area_vs_width_nbins',
        type=int, default=60,
        help='Number of bins for area vs width histogram for online monitor. '
             'NB: this is a 2D histogram'),
    strax.Option(
        'area_vs_width_bounds',
        type=tuple, default=((0, 5), (0, 5)),
        help='Boundaries of log-log histogram of area vs width'),
    strax.Option(
        'area_vs_width_cut_string',
        type=str, default='',
        help='Selection (like selection_str) applied to data for '
             '"area_vs_width_hist_clean", cuts should be separated using "&"'
             'For example: (tight_coincidence > 2) & (area_fraction_top < 0.1)'
             'Default is no selection (other than "area_vs_width_min_gap")'),
    strax.Option(
        'lone_hits_area_bounds',
        type=tuple, default=(0, 1500),
        help='Boundaries area histogram of lone hits [ADC]'),
    strax.Option(
        'online_peak_monitor_nbins',
        type=int, default=100,
        help='Number of bins of histogram of online monitor. Will be used '
             'for: '
             'lone_hits_area-histogram, '
             'area_fraction_top-histogram, '
             'online_se_gain estimate (histogram is not stored), '
    ),
    strax.Option(
        'lone_hits_cut_string',
        type=str,
        default='(area >= 50) & (area <= 250)',
        help='Selection (like selection_str) applied to data for '
             '"lone-hits", cuts should be separated using "&")'),
    strax.Option(
        'lone_hits_min_gap',
        type=int,
        default=15_000,
        help='Minimal gap [ns] between consecutive lone-hits. To turn off '
             'this cut, set to 0.'),
    strax.Option(
        'n_tpc_pmts', type=int,
        help='Number of TPC PMTs'),
    strax.Option(
        'online_se_bounds',
        type=tuple, default=(7, 70),
        help='Window for online monitor [PE] to look for the SE gain, value'
    )
)
class OnlinePeakMonitor(strax.Plugin):
    """
    Plugin to write data to the online-monitor. Data that is written by
    this plugin should be small such as to not overload the runs-
    database.

    This plugin takes 'peak_basics' and 'lone_hits'. Although they are
    not strictly related, they are aggregated into a single data_type
    in order to minimize the number of documents in the online monitor.

    Produces 'online_peak_monitor' with info on the lone-hits and peaks
    """
    depends_on = ('peak_basics', 'lone_hits')
    provides = 'online_peak_monitor'
    data_kind = 'online_peak_monitor'
    __version__ = '0.0.5'
    rechunk_on_save = False

    def infer_dtype(self):
        n_bins_area_width = self.config['area_vs_width_nbins']
        bounds_area_width = self.config['area_vs_width_bounds']

        n_bins = self.config['online_peak_monitor_nbins']

        n_tpc_pmts = self.config['n_tpc_pmts']
        dtype = [
            (('Start time of the chunk', 'time'),
             np.int64),
            (('End time of the chunk', 'endtime'),
             np.int64),
            (('Area vs width histogram (log-log)', 'area_vs_width_hist'),
             (np.int64, (n_bins_area_width, n_bins_area_width))),
            (('Area vs width edges (log-space)', 'area_vs_width_bounds'),
             (np.float64, np.shape(bounds_area_width))),
            (('Lone hits areas histogram [ADC-counts]', 'lone_hits_area_hist'),
             (np.int64, n_bins)),
            (('Lone hits areas bounds [ADC-counts]', 'lone_hits_area_bounds'),
             (np.float64, 2)),
            (('Lone hits per channel', 'lone_hits_per_channel'),
             (np.int64, n_tpc_pmts)),
            (('AFT histogram', 'aft_hist'),
             (np.int64, n_bins)),
            (('AFT bounds', 'aft_bounds'),
             (np.float64, 2)),
            (('Number of contributing channels histogram', 'n_channel_hist'),
             (np.int64, n_tpc_pmts)),
            (('Single electron gain', 'online_se_gain'),
             np.float32),
        ]
        return dtype

    def compute(self, peaks, lone_hits, start, end):
        # General setup
        res = np.zeros(1, dtype=self.dtype)
        res['time'] = start
        res['endtime'] = end
        n_pmt = self.config['n_tpc_pmts']
        n_bins = self.config['online_peak_monitor_nbins']

        # Bounds for histograms
        res['area_vs_width_bounds'] = self.config['area_vs_width_bounds']
        res['lone_hits_area_bounds'] = self.config['lone_hits_area_bounds']

        # -- Peak vs area 2D histogram --
        # Always cut out unphysical peaks
        sel = (peaks['area'] > 0) & (peaks['range_50p_area'] > 0)
        res['area_vs_width_hist'] = self.area_width_hist(peaks[sel])
        del sel

        # -- Lone hit properties --
        # Make a mask with the cuts.
        # NB: LONE HITS AREA ARE IN ADC!
        mask = _config_as_selection_str(
            self.config['lone_hits_cut_string'], lone_hits)
        # Now only take lone hits that are separated in time.
        lh_timedelta = lone_hits[1:]['time'] - strax.endtime(lone_hits)[:-1]
        # Hits on the left are far away? (assume first is because of chunk bound)
        mask &= np.hstack([True, lh_timedelta > self.config['lone_hits_min_gap']])
        # Hits on the right are far away? (assume last is because of chunk bound)
        mask &= np.hstack([lh_timedelta > self.config['lone_hits_min_gap'], True])

        # Make histogram of ADC counts
        lone_hit_areas, _ = np.histogram(lone_hits[mask]['area'],
                                         bins=n_bins,
                                         range=self.config['lone_hits_area_bounds'])

        lone_hit_channel_count, _ = np.histogram(lone_hits[mask]['channel'],
                                                 bins=n_pmt,
                                                 range=[0, n_pmt])
        # Count number of lone-hits per PMT
        res['lone_hits_area_hist'] = lone_hit_areas
        res['lone_hits_per_channel'] = lone_hit_channel_count
        # Clear mask, don't re-use
        del mask

        # -- AFT histogram --
        aft_b = [0, 1]
        aft_hist, _ = np.histogram(peaks['area_fraction_top'], bins=n_bins, range=aft_b)
        res['aft_hist'] = aft_hist
        res['aft_bounds'] = aft_b

        # Estimate Single Electron (SE) gain
        se_hist, se_bins = np.histogram(peaks['area'], bins=n_bins,
                                        range=self.config['online_se_bounds'])
        bin_centers = (se_bins[1:] + se_bins[:1]) / 2
        res['online_se_gain'] = bin_centers[np.argmax(se_hist)]
        return res

    def area_width_hist(self, data):
        """Make area vs width 2D-hist"""
        hist, _, _ = np.histogram2d(
            np.log10(data['area']),
            np.log10(data['range_50p_area']),
            range=self.config['area_vs_width_bounds'],
            bins=self.config['area_vs_width_nbins'])
        return hist.T


@export
@strax.takes_config(
    strax.Option(
        'drift_time_vs_R2_cut_string',
        type=str,
        default='(s2_index > -1) & (s1_index > -1)',
        help='Selection (like selection_str) applied to data for '
             '"drift_time_vs_R2", cuts should be separated using "&"'),
    strax.Option(
        'drift_time_vs_R2_bounds',
        type=tuple, default=((0, 5e3), (0, 3e6)),
        help='Boundaries of histogram of drift time vs R^2.'),
    strax.Option(
        'online_event_monitor_nbins',
        type=int, default=60,
        help='Number of bins of histogram of online monitor. NB: these are 2D '
             'histograms! Will be used for: '
             'drift_time_vs_R2-histogram, '
             'drift_time_vs_s2wdith-histogram, '
             'drift_time_vs_s1aft-histogram, '
             's1area_vs_s2area-histogram, '),
    strax.Option(
        'drift_time_vs_s2width_bounds',
        type=tuple, default=((0, 3e6), (0, 1.5e4)),
        help='Boundaries of histogram of drift time vs. s2width'),
    strax.Option(
        'drift_time_vs_s1aft_bounds',
        type=tuple, default=((0, 3e6), (0, 1)),
        help='Boundaries of histogram of drift time vs. s1aft'),
    strax.Option(
        's1area_vs_s2area_bounds',
        type=tuple, default=((0, 5), (1, 7)),
        help='Boundaries of the log-log histogram of s1 area vs. s2 area'),
)
class OnlineEventMonitor(strax.Plugin):
    """
    Plugin to write data to the online-monitor. Data that is written by
    this plugin should be small such as to not overload the runs-
    database.

    This plugin takes 'event_basics'.

    Produces 'online_event_monitor' with info on the events.
    """
    depends_on = ('event_basics',)
    provides = 'online_event_monitor'
    __version__ = '0.0.1'
    data_kind = 'online_event_monitor'
    rechunk_on_save = False

    def infer_dtype(self):
        n_bins = self.config['online_event_monitor_nbins']
        bounds_drift_time_R2 = self.config['drift_time_vs_R2_bounds']

        dtype = [
            (('Start time of the chunk', 'time'),
             np.int64),
            (('End time of the chunk', 'endtime'),
             np.int64),
            (('Drift time vs R^2 histogram', 'drift_time_vs_R2_hist'),
             (np.int64, (n_bins, n_bins))),
            (('Drift time vs R^2 edges (linear space)', 'drift_time_vs_R2_bounds'),
             (np.float64, np.shape(bounds_drift_time_R2))),
            (('Drift time vs. s2width histogram', 'drift_time_vs_s2width_hist'),
             (np.int64, (n_bins, n_bins))),
            (('Drift time vs. s2width edges', 'drift_time_vs_s2width_bounds'),
             (np.float64, np.shape(self.config['drift_time_vs_s2width_bounds']))),
            (('Drift time vs. s1aft histogram', 'drift_time_vs_s1aft_hist'),
             (np.int64, (n_bins, n_bins))),
            (('Drift time vs. s1aft edges', 'drift_time_vs_s1aft_bounds'),
             (np.float64, np.shape(self.config['drift_time_vs_s1aft_bounds']))),
            (('S1 area vs. S2 area histogram', 's1area_vs_s2area_hist'),
             (np.int64, (n_bins, n_bins))),
            (('S1 area vs. S2 area edges (log-log)', 's1area_vs_s2area_bounds'),
             (np.float64, np.shape(self.config['s1area_vs_s2area_bounds']))),
        ]
        return dtype

    def compute(self, events, start, end):
        # General setup
        res = np.zeros(1, dtype=self.dtype)
        res['time'] = start
        res['endtime'] = end

        # --- Drift time vs R^2 histogram ---
        res['drift_time_vs_R2_bounds'] = self.config['drift_time_vs_R2_bounds']

        mask = _config_as_selection_str(
            self.config['drift_time_vs_R2_cut_string'], events)
        res['drift_time_vs_R2_hist'] = self.drift_time_R2_hist(events[mask])
        # Make a new mask, don't re-use
        del mask

        # drift time vs. s2 width histogram
        res['drift_time_vs_s2width_bounds'] = self.config['drift_time_vs_s2width_bounds']
        res['drift_time_vs_s2width_hist'] = self.drift_time_s2width_hist(events)

        # drift time vs. s1 aft histogram
        res['drift_time_vs_s1aft_bounds'] = self.config['drift_time_vs_s1aft_bounds']
        res['drift_time_vs_s1aft_hist'] = self.drift_time_s1aft_hist(events)

        # s1 area vs. s2 area (log-log)
        res['s1area_vs_s2area_bounds'] = self.config['s1area_vs_s2area_bounds']
        res['s1area_vs_s2area_hist'] = self.s1area_s2area_hist(events)

        # Other event properties?
        return res

    def drift_time_R2_hist(self, data):
        """Make drift time vs R^2 2D-hist"""
        hist, _, _ = np.histogram2d(
            (data['s2_x'] ** 2 + data['s2_y'] ** 2),
            data['drift_time'],
            range=self.config['drift_time_vs_R2_bounds'],
            bins=self.config['online_event_monitor_nbins'])
        return hist.T

    def drift_time_s2width_hist(self, data):
        """Make drift time vs. s2width 2D-hist"""
        hist, _, _ = np.histogram2d(
            data['drift_time'],
            data['s2_range_50p_area'],
            range=self.config['drift_time_vs_s2width_bounds'],
            bins=self.config['online_event_monitor_nbins'])
        return hist.T

    def drift_time_s1aft_hist(self, data):
        """Make drift time vs. s1aft 2D-hist"""
        hist, _, _ = np.histogram2d(
            data['drift_time'],
            data['s1_area_fraction_top'],
            range=self.config['drift_time_vs_s1aft_bounds'],
            bins=self.config['online_event_monitor_nbins'])
        return hist.T

    def s1area_s2area_hist(self, data):
        """Make s1aft vs. drift time 2D-hist"""
        hist, _, _ = np.histogram2d(
            np.log10(data['s1_area']),
            np.log10(data['s2_area']),
            range=self.config['s1area_vs_s2area_bounds'],
            bins=self.config['online_event_monitor_nbins']
        )
        return hist.T


def _config_as_selection_str(selection_string, data, pre_sel=None):
    """Get mask for data base on the selection string"""
    if pre_sel is None:
        pre_sel = np.ones(len(data), dtype=np.bool_)

    if selection_string != '':
        if isinstance(selection_string, (list, tuple)):
            selection_string = ' & '.join(f'({x})' for x in selection_string)

        mask = numexpr.evaluate(selection_string, local_dict={
            fn: data[fn]
            for fn in data.dtype.names})
        pre_sel &= mask
    return pre_sel



import bokeh
import bokeh.plotting as bklt
import numpy as np
import strax
import straxen

# Default legend and color options for unknow, S1 and S2
COLORS = ('gray', 'blue', 'green')
LEGENDS = ('Unknown', 'S1', 'S2')


@straxen.mini_analysis(requires=('events', 'event_basics', 'peaks', 'peak_basics', 'peak_positions'), 
                       warn_beyond_sec=0.05)
def event_display_interactive(events, peaks, to_pe, run_id, context, xenon1t=False, log=True):
    """
    Interactive event display for XENONnT. Plots detailed main/alt
    S1/S2, bottom and top PMT hit pattern as well as all other peaks
    in a given event.
    
    :param xenin1T: Flag to use event display with 1T data.
    :param log: If true color sclae is used for hitpattern plots.

    Note:
        How to use:

        > st.event_display(context,
        >                  run_id,
        >                  time_range=(event['time'],
        >                              event['endtime'])
        >                  )

    Warning:
        Raises an error if the user queries a time range which contains
        more than a single event.

    :return: bokeh.plotting.figure instance.
    """
    if len(events) != 1:
        raise ValueError('The time range you specified contains more or'
                         ' less than a single event. The event display '
                         ' only works with individual events for now.')
        
    if peaks.shape[0] == 0:
        raise ValueError('Found an event without peaks this should not had have happened.')
    
    # Select main/alt S1/S2s based on time and endtime in event:
    m_other_peaks = np.ones(len(peaks), dtype=np.bool_)  # To select non-event peaks
    endtime = strax.endtime(peaks)

    signal = {}
    keys = ['s1', 'alt_s1', 's2', 'alt_s2']
    for s in keys:
        m = (peaks['time'] == events[f'{s}_time']) & (endtime == events[f'{s}_endtime'])
        signal[s] = peaks[m]
        m_other_peaks &= ~m
    
    # Use consistent labels:
    labels = {'s1': 'MS1', 'alt_s1': 'AS1', 's2': 'MS2', 'alt_s2': 'AS2'}

    # Detail plot S1/S2:
    fig_s1 = straxen.bokeh_utils.default_fig(title='Main/Alt S1')
    fig_s2 = straxen.bokeh_utils.default_fig(title='Main/Alt S2')
    for fig, peak_types in zip([fig_s1, fig_s2], (keys[:2], keys[2:])):
        pi = 0
        for peak_type in peak_types:
            p = None
            if 's2' in peak_type:
                time_scalar = 1000
            else:
                time_scalar = 1
            if signal[peak_type].shape[0]:
                fig, p = plot_peak_detail(signal[peak_type], 
                                          time_scaler=time_scalar,
                                          label=labels[peak_type], 
                                          fig=fig)
            if pi and p:
                # Not main S1/S2
                p.visible = False
            pi += 1
            
    # PMT arrays:
    # TOP
    fig_top = straxen.bokeh_utils.default_fig(title='top array')
    fig_bottom = straxen.bokeh_utils.default_fig(title='bottom array')
    pmt_arrays = {'top': (fig_top, keys[2:] + keys[:2]),
                  'bottom': (fig_bottom, keys)}
    for parray, v in pmt_arrays.items():
        fig, peak_type = v

        for ind, k in enumerate(peak_type):
            if not signal[k].shape[0]:
                # alt S1/S2 does not exist
                continue

            fig, p, _ = plot_pmt_array(signal[k][0], parray, to_pe,
                                       label=labels[k], xenon1t=xenon1t, fig=fig, 
                                       log=log)
            if ind:
                # Not main S1 or S2
                p.visible = False

            if parray == 'top' and 's2' in k:
                # In case of the top PMT array we also have to plot the S2 positions:
                fig, p = plot_posS2s(signal[k][0], label=labels[k], fig=fig, s2_type_style_id=ind)
                if ind:
                    # Not main S2
                    p.visible = False

    # Now we only have to add all other S2 to the top pmt array
    m_other_s2 = m_other_peaks & (peaks['type'] == 2)
    if np.any(m_other_s2):
        fig_top, p = plot_posS2s(peaks[m_other_s2], label='OS2s', fig=fig_top, s2_type_style_id=2)
        p.visible = False

    # Main Plot:
    title = _make_event_title(events[0], run_id)
    waveform = plot_event(peaks, signal, labels)

    # Put everything together:
    upper_row = bokeh.layouts.Row(children=[fig_s1, fig_s2, fig_top, fig_bottom])

    plots = bokeh.layouts.gridplot(children=[upper_row, waveform],
                                           sizing_mode='scale_both',
                                           ncols=1,
                                           merge_tools=True,
                                           toolbar_location='above',
                                           )
    event_display = bokeh.layouts.Column(children=[title, plots],
                                         sizing_mode='scale_both',
                                         max_width=1600,
                                        )
    

    return event_display


def plot_event(peaks, signal, labels):
    """
    Wrapper for plot peaks to highlight main/alt. S1/S2

    :param peaks: Peaks in event
    :param signal: Dictionary containing main/alt. S1/S2
    :param labels: dict with labels to be used
    :return: bokeh.plotting.figure instance
    """
    waveform = plot_peaks(peaks, time_scaler=1000)
    # Hightlight main and alternate S1/S2:
    start = peaks[0]['time']
    # Workaround did not manage to scale via pixels...
    ymax = np.max((peaks['data'].T/peaks['dt']).T)
    ymax -= 0.1 * ymax
    for s, p in signal.items():
        if p.shape[0]:
            pos = (p[0]['center_time'] - start) / 1000
            main = bokeh.models.Span(location=pos,
                                     dimension='height',
                                     line_alpha=0.6,
                                     )
            vline_label = bokeh.models.Label(x=pos,
                                             y=ymax,
                                             angle=np.pi / 2,
                                             text=labels[s],
                                             )
            if 'alt' in s:
                main.line_dash = 'dotted'
            else:
                main.line_dash = 'dashed'
            waveform.add_layout(main)
            waveform.add_layout(vline_label)
    return waveform


def plot_peak_detail(peak, time_scaler=1, label='', fig=None):
    """
    Function which makes a detailed plot for the given peak. As in the
    main/alt S1/S2 plots of the event display.

    :param peak: Peak to be plotted.
    :param time_scaler: Factor to rescale the time from ns to other scale.
        E.g. =1000 scales to µs.
    :param label: Label to be used in the plot legend.
    :param fig: Instance of bokeh.plotting.figure if None one will be
        created via straxen.bokeh.utils.default_figure().
    :return: Instance of bokeh.plotting.figure
    """
    if not peak.shape:
        peak = np.array([peak])

    if peak.shape[0] != 1:
        raise ValueError('Cannot plot the peak details for more than one '
                         'peak. Please make sure peaks has the shape (1,)!')

    p_type = peak[0]['type']

    if not fig:
        fig = straxen.bokeh_utils.default_fig(title=f'Main/Alt S{p_type}')

    tt = straxen.bokeh_utils.peak_tool_tip(p_type)
    tt = [v for k, v in tt.items() if k not in ['time_static', 'center_time', 'endtime']]
    fig.add_tools(bokeh.models.HoverTool(names=[label],
                                         tooltips=tt)
                  )

    source = straxen.bokeh_utils.get_peaks_source(peak,
                                                  relative_start=peak[0]['time'],
                                                  time_scaler=time_scaler,
                                                  keep_amplitude_per_sample=False
                                                  )

    patches = fig.patches(source=source,
                          legend_label=label,
                          fill_color=COLORS[p_type],
                          fill_alpha=0.2,
                          line_color=COLORS[p_type],
                          line_width=0.5,
                          name=label
                          )
    fig.xaxis.axis_label = f"Time [{time_scaler} ns]"
    fig.xaxis.axis_label_text_font_size = '12pt'
    fig.yaxis.axis_label = "Amplitude [pe/ns]"
    fig.yaxis.axis_label_text_font_size = '12pt'

    fig.legend.location = "top_right"
    fig.legend.click_policy = "hide"

    if label:
        fig.legend.visible = True
    else:
        fig.legend.visible = False

    return fig, patches


def plot_peaks(peaks, time_scaler=1, fig=None):
    """
    Function which plots a list/array of peaks relative to the first
    one.

    :param peaks: Peaks to be plotted.
    :param time_scaler:  Factor to rescale the time from ns to other scale.
        E.g. =1000 scales to µs.
    :param fig: Instance of bokeh.plotting.figure if None one will be
        created via straxen.bokeh.utils.default_figure().
    :return: bokeh.plotting.figure instance.
    """
    if not fig:
        fig = straxen.bokeh_utils.default_fig(width=1600, height=400)

    for i in range(0, 3):
        _ind = np.where(peaks['type'] == i)[0]
        if not len(_ind):
            continue
        
        source = straxen.bokeh_utils.get_peaks_source(peaks[_ind],
                                                      relative_start=peaks[0]['time'],
                                                      time_scaler=time_scaler,
                                                      keep_amplitude_per_sample=False
                                                      )

        fig.patches(source=source,
                    fill_color=COLORS[i],
                    fill_alpha=0.2,
                    line_color=COLORS[i],
                    line_width=0.5,
                    legend_label=LEGENDS[i],
                    name=LEGENDS[i],
                    )

        tt = straxen.bokeh_utils.peak_tool_tip(i)
        tt = [v for k, v in tt.items() if k != 'time_dynamic']
        fig.add_tools(bokeh.models.HoverTool(names=[LEGENDS[i]], tooltips=tt))

    fig.xaxis.axis_label = 'Time [µs]'
    fig.xaxis.axis_label_text_font_size = '14pt'
    fig.yaxis.axis_label = "Amplitude [pe/ns]"
    fig.yaxis.axis_label_text_font_size = '14pt'

    fig.legend.location = "top_left"
    fig.legend.click_policy = "hide"
    return fig


def plot_pmt_array(peak, array_type, to_pe, log=False, xenon1t=False, fig=None, label=''):
    """
    Plots top or bottom PMT array for given peak.

    :param peak: Peak for which the hit pattern should be plotted.
    :param array_type: String which specifies if "top" or "bottom"
        PMT array should be plotted
    :param to_pe: PMT gains.
    :param log: If true use a log-scale for the colorscale.
    :param fig: Instance of bokeh.plotting.figure if None one will be
        created via straxen.bokeh.utils.default_figure().
    :param label: Label of the peak which should be used for the
        plot legend

    :returns: Tuple containing a bokeh figure, glyph and transform
        instance.
    """
    if peak.shape:
        raise ValueError('Can plot PMT array only for a single peak at a time.')

    tool_tip = [('Plot', '$name'),
                ('Channel', '@pmt'),
                ('X-Position [cm]', '$x'),
                ('Y-Position [cm]', '$y'),
                ('area [pe]', '@area')
                ]

    array = ('top', 'bottom')
    if array_type not in array:
        raise ValueError('"array_type" must be either top or bottom.')

    if not fig:
        fig = straxen.bokeh_utils.default_fig(title=f'{array_type} array')

    # Creating TPC axis and title
    fig = _plot_tpc(fig)

    # Plotting PMTs:
    pmts = straxen.pmt_positions(xenon1t)
    mask_pmts = to_pe == 0
    pmts_on = pmts[~mask_pmts]
    pmts_on = pmts_on[pmts_on['array'] == array_type]

    if np.any(mask_pmts):
        pmts_off = pmts[mask_pmts]
        pmts_off = pmts_off[pmts_off['array'] == array_type]
        fig = _plot_off_pmts(pmts_off, fig)

    area_per_channel = peak['area_per_channel'][pmts_on['i']]
    
    if log==True:
        area_plot = np.log10(area_per_channel)
        # Manually set infs to zero since cmap cannot handle it.
        area_plot = np.where(area_plot==-np.inf, 0, area_plot)  
    else:
        area_plot = area_per_channel
        
    mapper = bokeh.transform.linear_cmap(field_name='area_plot',
                                         palette="Viridis256",
                                         low=min(area_plot),
                                         high=max(area_plot))

    source_on = bklt.ColumnDataSource(data={'x': pmts_on['x'],
                                            'y': pmts_on['y'],
                                            'area': area_per_channel,
                                            'area_plot': area_plot,
                                            'pmt': pmts_on['i']
                                            }
                                      )

    p = fig.scatter(source=source_on,
                    radius=straxen.tpc_pmt_radius,
                    fill_color=mapper,
                    fill_alpha=1,
                    line_color='black',
                    legend_label=label,
                    name=label + '_pmt_array',
                    )
    fig.add_tools(bokeh.models.HoverTool(names=[label + '_pmt_array'], tooltips=tool_tip))
    fig.legend.location = 'top_left'
    fig.legend.click_policy = "hide"
    fig.legend.orientation = "horizontal"
    fig.legend.padding = 0
    fig.toolbar_location = None
    return fig, p, mapper


def _plot_tpc(fig=None):
    """
    Plots ring at TPC radius and sets xy limits + labels.
    """
    if not fig:
        fig = straxen.bokeh_utils.default_fig()

    fig.circle(x=0, y=0,
               radius=straxen.tpc_r,
               fill_color="white",
               line_color="black",
               line_width=3,
               fill_alpha=0
               )
    fig.xaxis.axis_label = 'x [cm]'
    fig.xaxis.axis_label_text_font_size = '12pt'
    fig.yaxis.axis_label = 'y [cm]'
    fig.yaxis.axis_label_text_font_size = '12pt'
    fig.x_range.start = -80
    fig.x_range.end = 80
    fig.y_range.start = -80
    fig.y_range.end = 80

    return fig


def _plot_off_pmts(pmts, fig=None):
    """
    Plots PMTs which are switched off.
    """
    if not fig:
        fig = straxen.bokeh_utils.default_fig()
    fig.circle(x=pmts['x'],
               y=pmts['y'],
               fill_color='gray',
               line_color='black',
               radius=straxen.tpc_pmt_radius,
               )
    return fig


def plot_posS2s(peaks, label='', fig=None, s2_type_style_id=0):
    """
    Plots xy-positions of specified peaks.

    :param peaks: Peaks for which the position should be plotted.
    :param label: Legend label and plot name (name serves as idenitfier).
    :param fig: bokeh.plotting.figure instance the plot should be plotted
        into. If None creates new instance.
    :param s2_type_style_id: 0 plots main S2 style, 1 for alt S2 and
        2 for other S2s (e.g. single electrons).
    """
    if not peaks.shape:
        peaks = np.array([peaks])
    
    if not np.all(peaks['type'] == 2):
        raise ValueError('All peaks must be S2!')

    if not fig:
        fig = straxen.bokeh_utils.default_fig()

    source = straxen.bokeh_utils.get_peaks_source(peaks)

    if s2_type_style_id == 0:
        p = fig.cross(source=source,
                      name=label,
                      legend_label=label,
                      color='red',
                      line_width=2,
                      size=12)

    if s2_type_style_id == 1:
        p = fig.cross(source=source,
                      name=label,
                      legend_label=label,
                      color='orange',
                      angle=45 / 360 * 2 * np.pi,
                      line_width=2,
                      size=12)

    if s2_type_style_id == 2:
        p = fig.diamond_cross(source=source,
                              name=label,
                              legend_label=label,
                              color='red',
                              size=8
                              )

    tt = straxen.bokeh_utils.peak_tool_tip(2)
    tt = [v for k, v in tt.items() if k not in ['time_dynamic', 'amplitude']]
    fig.add_tools(bokeh.models.HoverTool(names=[label],
                                         tooltips=[('Position x [cm]', '@x'),
                                                   ('Position y [cm]', '@y')] + tt))
    return fig, p


def _make_event_title(event, run_id, width=1600):
    """
    Function which makes the title of the plot for the specified event.

    Note:
        To center the title I use a transparent box.

    :param event: Event which we are plotting
    :param run_id: run_id

    :returns: Title as bokeh.models.Div instance
    """
    start = event['time']
    date = np.datetime_as_string(start.astype('<M8[ns]'), unit='s')
    start_ns = start - (start // 10**9) * 10**9
    end = strax.endtime(event)
    end_ns = end - start + start_ns
    event_number = event['event_number']
    text = (f'<h2>Event {event_number} from run {run_id}<br>'
            f'Recorded at {date[:10]} {date[10:]} UTC,'
            f' {start_ns} ns - {end_ns} ns </h2>')

    title = bokeh.models.Div(text=text,
                             style={'text-align': 'left',
                                    },
                             sizing_mode='scale_both',
                             width = width,
                             default_size=width,
                             orientation='vertical',
                             width_policy='fit',
                             margin=(0,0,-30, 50)
                             )
    return title



"""
Test certain interactions with the runsdatabase.
NB! this only works if one has access to the database. This does not
work e.g. on travis jobs and therefore the tests failing locally will
not show up in Pull Requests.
"""

import straxen
import os
from warnings import warn
from .test_plugins import test_run_id_nT


def test_select_runs(check_n_runs=2):
    """
    Test (if we have a connection) if we can perform strax.select_runs
        on the last two runs in the runs collection

    :param check_n_runs: int, the number of runs we want to check
    """

    if not straxen.utilix_is_configured():
        warn('Makes no sense to test the select runs because we do not '
             'have access to the database.')
        return
    assert check_n_runs >= 1
    st = straxen.contexts.xenonnt_online()
    run_col = st.storage[0].collection

    # Find the latest run in the runs collection
    last_run = run_col.find_one(projection={'number': 1},
                                sort=[('number', -1)]
                                ).get('number')

    # Set this number as the minimum run number. This limits the
    # amount of documents checked and therefore keeps the test short.
    st.storage[0].minimum_run_number = int(last_run) - (check_n_runs - 1)
    st.select_runs()


def test_downloader():
    """Test if we can download a small file from the downloader"""
    if not straxen.utilix_is_configured():
        warn('Cannot download because utilix is not configured')
        return

    downloader = straxen.MongoDownloader()
    path = downloader.download_single('to_pe_nt.npy')
    assert os.path.exists(path)


def test_elife():
    """Test the get_elife function from CMT"""
    if not straxen.utilix_is_configured():
        warn('Cannot test CMT elife since utilix is not configured')
        return
    cmt = straxen.CorrectionsManagementServices()
    cmt.get_elife(test_run_id_nT, 'elife_model', 'ONLINE')


def _patch_om_init(take_only):
    """
    temp patch since om = straxen.OnlineMonitor() does not work with utilix
    """
    header = 'RunDB'
    user = straxen.uconfig.get(header, 'pymongo_user')
    pwd = straxen.uconfig.get(header, 'pymongo_password')
    url = straxen.uconfig.get(header, 'pymongo_url').split(',')[-1]
    uri = f"mongodb://{user}:{pwd}@{url}"
    return straxen.OnlineMonitor(uri=uri, take_only=take_only)


def test_online_monitor(target='online_peak_monitor', max_tries=3):
    """
    See if we can get some data from the online monitor before max_tries

    :param target: target to test getting from the online monitor
    :param max_tries: number of queries max allowed to get a non-failing
        run
    """
    if not straxen.utilix_is_configured():
        warn('Cannot test online monitor because utilix is not configured')
        return
    st = straxen.contexts.xenonnt_online()
    om = _patch_om_init(target)
    st.storage = [om]
    for i in range(max_tries):
        some_run = om.db[om.col_name].find_one({'provides_meta': True,
                                                'data_type': target,
                                                },
                                               projection={'number': 1,
                                                           'metadata': 1})
        if some_run is None or some_run.get('number', None) is None:
            continue
        elif 'exception' in some_run.get('metadata', {}):
            continue
        else:
            # Correctly written
            run_id = f'{some_run["number"]:06}'
            break
    else:
        raise FileNotFoundError(f'No non-failing {target} found in the online '
                                f'monitor after {max_tries}')
    st.get_array(run_id, target, seconds_range=(0, 1), allow_incomplete=True)
